# -*- coding: utf-8 -*-
import numpy as np
from numpy import swapaxes as swap
from typing import Union, Iterable

from linkeddeepdict import LinkedDeepDict
from neumann.array import atleast1d

from .problem import NavierProblem
from .loads import LoadGroup, NavierLoadError
from .preproc import lhs_Navier, rhs_Bernoulli
from .postproc import postproc
from .proc import linsolve_Bernoulli, linsolve_Timoshenko


class NavierBeam(NavierProblem):
    """
    A class to handle simply-supported plates and their solution
    using Navier's approach.

    Parameters
    ----------
    length : float
        The length of the beam.

    N : int, Optional
        The number of harmonic terms involved in the approximation.
        Default is 100.

    EI : float
        Bending stiffness. Default is None.

    GA : float, Optional
        Shear stiffness. Only for Timoshenko beams. Default is None.

    Examples
    --------
    To define an Euler-Bernoulli beam of length 10.0 and 
    bending stiffness 2000.0 with 100 harmonic terms involved:
    
    >>> from sigmaepsilon.solid.fourier import NavierBeam
    >>> beam = NavierBeam(10.0, 100, EI=2000.0)
    
    To define a Timoshenko beam of length 10.0, bending stiffness 
    2000.0 and shear stiffness 1500.0 with 100 harmonic terms involved:
    
    >>> from sigmaepsilon.solid.fourier import NavierBeam
    >>> beam = NavierBeam(10.0, 100, EI=2000.0, GA=1500.0)
    
    """

    def __init__(self, length: float, N: int = 100, *,
                 EI: float = None, GA: float = None):
        super().__init__()
        self.length = length
        self.EI = EI
        self.GA = GA
        self.N = N

    def solve(self, loads: Union[dict, LoadGroup], 
              points: Union[float, Iterable]) -> LinkedDeepDict:
        """
        Solves the problem and calculates all postprocessing quantities at 
        one ore more points.

        Parameters
        ----------
        loads : Union[dict, LoadGroup]
            The loads. If it is an array, it should be a 2d float array.

        points : float or Iterable
            A float or an 1d iterable of coordinates, where the results are 
            to be evaluated. If it is a scalar, the resulting dictionary
            contains 1d arrays for every quantity, for every load case. If 
            there are multiple points, the result attached to a load case is 
            a 2d array, where the first axis goes along the points.

        Returns
        -------
        LinkedDeepDict
            A nested dictionary with a same layout as the input.

        """
        # STIFFNESS
        LHS = lhs_Navier(self.length, self.N, D=self.EI,
                         S=self.GA, squeeze=False)

        # LOADS
        if isinstance(loads, LoadGroup):
            _loads = loads
        elif isinstance(loads, dict):
            _loads = LoadGroup.from_dict(loads)
        else:
            raise NavierLoadError()
        _loads.problem = self
        LC = list(_loads.cases())
        RHS = np.stack(list(lc.rhs() for lc in LC))

        # SOLUTION
        if self.GA is None:
            _RHS = rhs_Bernoulli(RHS, self.length)
            coeffs = linsolve_Bernoulli(LHS, _RHS)
            del _RHS
            # (nRHS, nLHS, nMN)
        else:
            coeffs = linsolve_Timoshenko(LHS, RHS)
            # (nRHS, nLHS, nMN, 2)

        # POSTPROCESSING
        points = atleast1d(points)
        res = postproc(self.length, self.N, points, coeffs,
                       RHS, self.EI, self.GA, squeeze=False)
        # (nRHS, nLHS, nP, nX)
        res = swap(res, 1, 2)
        # (nRHS, nP, nLHS, nX)
        result = LinkedDeepDict()
        for i, lc in enumerate(LC):
            result[lc.address] = np.squeeze(res[i, :, :])
        result.lock()
        return result



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from numba import njit, prange

from polymesh.cells import T3 as Triangle
from polymesh.tri.triutils import lcoords_tri, shp_tri_loc

from ..model.membrane import Membrane

from .elem import FiniteElement
from .metaelem import ABCFiniteElement as ABC

__cache = True


__all__ = ['LSTMR']


def sig_LSTMR(key: str = 'opt', C: ndarray = None):
    if key == 'opt':
        sig = sig_LSTMR_opt(C)
    elif key == 'ALL3I':
        sig = (1, 4/9, 1/12, 5/12, 1/2, 0, 1/3, -1/3, -1/12, -1/2, -5/12)
    elif key == 'ALL3M':
        sig = (1, 4/9, 1/4, 5/4, 3/2, 0, 1, -1, -1/4, -3/2, -5/4)
    elif key == 'ALLLS':
        sig = (1, 4/9, 3/20, 3/4, 9/10, 0, 3/5, -3/5, -3/20, -9/10, -3/4)
    elif key == 'LSTRet':
        sig = (4/3, 1/2, 2/3, -2/3, 0, 0, -4/3, 4/3, -2/3, 0, 2/3)
    elif key == 'CST':
        sig = (0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0)
    else:
        raise NotImplementedError
    return np.array(sig).astype(float)


@njit(nogil=True, cache=__cache)
def sig_LSTMR_opt(C: ndarray):
    E11, E12, E13, _, E22, E23, _, _, E33 = C.flatten()
    Edet = E11*E22*E33 + 2*E12*E13*E23 - E11*E23**2 - \
        E22*E13**2 - E33*E12**2
    W = -6*E12**3 + 5*E11**2*E22 - 5*E12**2*E22 - \
        E22*(75*E13**2 + 14*E13*E23 + 3*E23**2) + \
        2*E12*(7*E13**2 + 46*E13*E23 + 7*E23**2) - \
        E11*(5*E12**2+3*E13**2 - 6*E12*E22 - 5*E22**2 + 14*E13*E23 +
             75*E23**2) + (3*E11**2 + 82*E11*E22 + 3*E22**2 -
                           4*(6*E12**2 + 5*E13**2 -
                              6*E13*E23+5*E23**2))*E33 \
        + 4*(5*E11-6*E12+5*E22)*E33**2
    E11C11avg = W / (128 * Edet)
    b0 = max([2/E11C11avg - 3/2, 1/100])
    return np.array([3/2, b0, 1, 2, 1, 0, 1, -1, -1, -1, -2], dtype=C.dtype)


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def sig_LSTMR_opt_bulk(C: ndarray):
    nE = len(C)
    res = np.zeros((nE, 11), dtype=C.dtype)
    for i in prange(nE):
        res[i] = sig_LSTMR_opt(C[i])
    return res


@njit(nogil=True, cache=__cache)
def _stiffness_data_(ecoords, h, sig):
    (x1, x2, x3), (y1, y2, y3) = ecoords[:, 0], ecoords[:, 1]
    ab, b0, b1, b2, b3, b4, b5, b6, b7, b8, b9 = sig
    x12, x21, x13, x31, x23, x32 = \
        x1 - x2, x2 - x1, x1 - x3, x3 - x1, x2 - x3, x3 - x2
    y12, y21, y13, y31, y23, y32 = \
        y1 - y2, y2 - y1, y1 - y3, y3 - y1, y2 - y3, y3 - y2
    LL21 = x21**2 + y21**2
    LL32 = x32**2 + y32**2
    LL13 = x13**2 + y13**2
    A = (y21 * x13 - x21 * y13)/2
    A2, A4 = 2*A, 4*A
    L = np.array([
        [y23, 0, x32],
        [0, x32, y23],
        [ab*y23*(y13-y21)/6, ab*x32*(x31-x12)/6, ab*(x31*y13-x12*y21)/3],
        [y31, 0, x13],
        [0, x13, y31],
        [ab*y31*(y21-y32)/6, ab*x13*(x12-x23)/6, ab*(x12*y21-x23*y32)/3],
        [y12, 0, x21],
        [0, x21, y12],
        [ab*y12*(y32-y13)/6, ab*x21*(x23-x31)/6, ab*(x23*y32-x31*y13)/3],
    ], dtype=ecoords.dtype) * h / 2
    Tru = np.array([
        [x32, y32, A4, x13, y13, 0,  x21, y21, 0],
        [x32, y32, 0,  x13, y13, A4, x21, y21, 0],
        [x32, y32, 0,  x13, y13, 0,  x21, y21, A4]
    ], dtype=ecoords.dtype) / A4
    Te = np.array([
        [y23*y13*LL21, y31*y21*LL32, y12*y32*LL13],
        [x23*x13*LL21, x31*x21*LL32, x12*x32*LL13],
        [(y23*x31+x32*y13)*LL21, (y31*x12+x13*y21)*LL32,
         (y12*x23+x21*y32)*LL13]
    ], dtype=ecoords.dtype) / (4 * A**2)
    Q1 = np.array([
        [b1/LL21, b2/LL21, b3/LL21],
        [b4/LL32, b5/LL32, b6/LL32],
        [b7/LL13, b8/LL13, b9/LL13]
    ], dtype=ecoords.dtype) * A2 / 3
    Q2 = np.array([
        [b9/LL21, b7/LL21, b8/LL21],
        [b3/LL32, b1/LL32, b2/LL32],
        [b6/LL13, b4/LL13, b5/LL13]
    ], dtype=ecoords.dtype) * A2 / 3
    Q3 = np.array([
        [b5/LL21, b6/LL21, b4/LL21],
        [b8/LL32, b9/LL32, b7/LL32],
        [b2/LL13, b3/LL13, b1/LL13]
    ], dtype=ecoords.dtype) * A2 / 3
    return A, L, Te, b0, Q1, Q2, Q3, Tru


@njit(nogil=True, cache=__cache)
def _stiffness_matrix_LSTMR(C, ecoords, h, sig):
    A, L, Te, b0, Q1, Q2, Q3, Tru = _stiffness_data_(ecoords, h, sig)
    V = h * A
    Q4, Q5, Q6 = (Q1 + Q2) / 2, (Q3 + Q2) / 2, (Q1 + Q3) / 2
    Kb = (L @ C @ L.T) / V
    Cn = Te.T @ (C @ Te)
    Kr = (3 * b0 * V / 4) * \
        (Q4.T @ Cn @ Q4 + Q5.T @ Cn @ Q5 + Q6.T @ Cn @ Q6)
    Kh = Tru.T @ Kr @ Tru
    return Kb, Kh


@njit(nogil=True, cache=__cache)
def __stiffness_matrix__(C, ecoords, h, sig):
    Kb, Kh = _stiffness_matrix_LSTMR(C, ecoords, h, sig)
    return Kb + Kh


@njit(nogil=True, parallel=True, cache=__cache)
def stiffness_matrix(C, ecoords, h, sig):
    nE = len(C)
    res = np.zeros((nE, 9, 9), dtype=C.dtype)
    for i in prange(nE):
        res[i, :, :] = __stiffness_matrix__(C[i], ecoords[i], h[i], sig[i])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def stiffness_matrix_LSQ_4_12(C: ndarray, ecoords: ndarray,
                              h: ndarray, sig: ndarray,
                              dmaps: ndarray, subtopo: ndarray):
    nE = len(C)
    dmax = dmaps.max() + 1
    res = np.zeros((nE, dmax, dmax), dtype=C.dtype)
    for i in range(subtopo.shape[0]):
        K_i_bulk = stiffness_matrix(C, ecoords[:, subtopo[i], :], h, sig)
        for j in prange(K_i_bulk.shape[1]):
            for k in prange(K_i_bulk.shape[2]):
                res[:, dmaps[i, j], dmaps[i, k]] += K_i_bulk[:, j, k]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def approx_strains(pcoord, ecoords, h, sig, dofsol1d, gnum):
    nE = len(gnum)
    res = np.zeros((nE, 3), dtype=ecoords.dtype)
    A1, A2, A3 = shp_tri_loc(pcoord)
    for i in prange(nE):
        A, L, Te, _, Q1, Q2, Q3, Tru = \
            _stiffness_data_(ecoords[i], h[i], sig[i])
        uE = dofsol1d[gnum]
        #uE = element_dof_solution_bulk(dofsol1d, gnum[i], dtype)
        res[i, :] = (1 / (h[i] * A)) * (L.T @ uE) + \
            3 * (Te @ (A1*Q1 + A2*Q2 + A3*Q3) @ Tru @ uE) / 2
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def approx_stresses(pcoord, C, ecoords, h, sig, dofsol1d, gnum):
    e = approx_strains(pcoord, ecoords, h, sig, dofsol1d, gnum)
    nE = len(e)
    res = np.zeros((nE, 3), dtype=C.dtype)
    for i in prange(nE):
        res[i, :] = C[i] @ e[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def stresses_at_nodes(C, ecoords, h, sig, dofsol1d, gnum):
    lcoords = lcoords_tri()
    nE = len(C)
    res = np.zeros((nE, 3, 3), dtype=C.dtype)
    for i in prange(3):
        res[:, i, :] = approx_stresses(lcoords[i], C, ecoords, h,
                                       sig, dofsol1d, gnum)
    return res


class LSTMR(ABC, Membrane, Triangle, FiniteElement):
    """
    A template for 3-noded linear strain triangles with drilling degrees
    of freedom. This element is currently only applicable to one layer with an isotropic
    material definition.
    
    """
    __sig__ = 'opt'



# -*- coding: utf-8 -*-
import numpy as np
from numba import njit, prange
from collections import Iterable

from neumann.array import (ascont, atleast1d, atleast2d,
                           atleast3d, atleast4d)

from ..utils import layers_of_points, points_of_layers
from .utils import (z_to_shear_factors, shear_correction_data,
                    material_stiffness_matrices)

__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def _pproc_Mindlin_3D(ABDS: np.ndarray, sfx: np.ndarray, sfy: np.ndarray,
                      C_126: np.ndarray, C_45: np.ndarray,
                      bounds: np.ndarray, points: np.ndarray,
                      e_126_n: np.ndarray, c_126: np.ndarray,
                      e_45: np.ndarray):
    nRHS, nP, _ = e_126_n.shape
    layerinds = layers_of_points(points, bounds)

    # results
    s_126_n = np.zeros((nRHS, nP, 3), dtype=ABDS.dtype)
    s_126_m = np.zeros((nRHS, nP, 3), dtype=ABDS.dtype)
    s_45 = np.zeros((nRHS, nP, 2), dtype=ABDS.dtype)
    e_126_m = np.zeros((nRHS, nP, 3), dtype=ABDS.dtype)
    e_45_new = np.zeros((nRHS, nP, 2), dtype=ABDS.dtype)

    for iP in prange(nP):
        zP = points[iP, 2]
        lP = layerinds[iP]
        sfxz, sfyz = z_to_shear_factors(zP, sfx[lP, :], sfy[lP, :])
        for iRHS in prange(nRHS):
            e_126_m[iRHS, iP] = zP * c_126[iRHS, iP]
            s_126_n[iRHS, iP] = C_126[lP] @ e_126_n[iRHS, iP]
            s_126_m[iRHS, iP] = C_126[lP] @ e_126_m[iRHS, iP]
            s_45[iRHS, iP, 0] = sfxz * ABDS[6, 6] * e_45[iRHS, iP, 0]
            s_45[iRHS, iP, 1] = sfyz * ABDS[7, 7] * e_45[iRHS, iP, 1]
            e_45_new[iRHS, iP, 0] = s_45[iRHS, iP, 0] / C_45[lP, 0, 0]
            e_45_new[iRHS, iP, 1] = s_45[iRHS, iP, 1] / C_45[lP, 1, 1]

    return s_126_n, s_126_m, s_45, e_126_n, e_126_m, e_45_new


@njit(nogil=True, parallel=True, cache=__cache)
def _pproc_Mindlin_3D_s(ABDS: np.ndarray, sfx: np.ndarray, sfy: np.ndarray,
                        C_126: np.ndarray, C_45: np.ndarray,
                        bounds: np.ndarray, ppl: int,
                        e_126_n: np.ndarray, c_126: np.ndarray,
                        e_45: np.ndarray):
    nRHS, nP, _ = e_126_n.shape
    z = points_of_layers(bounds, ppl)
    nL = z.shape[0]

    # results
    s_126_n = np.zeros((nRHS, nP, nL, ppl, 3), dtype=ABDS.dtype)
    s_126_m = np.zeros((nRHS, nP, nL, ppl, 3), dtype=ABDS.dtype)
    s_45 = np.zeros((nRHS, nP, nL, ppl, 2), dtype=ABDS.dtype)
    e_126_m = np.zeros((nRHS, nP, nL, ppl, 3), dtype=ABDS.dtype)
    e_45_new = np.zeros((nRHS, nP, nL, ppl, 2), dtype=ABDS.dtype)

    for iL in prange(nL):
        for ippl in prange(ppl):
            zP = z[iL, ippl]
            sfxz, sfyz = z_to_shear_factors(zP, sfx[iL, :], sfy[iL, :])
            for iP in prange(nP):
                for iRHS in prange(nRHS):
                    e_126_m[iRHS, iP, iL, ippl] = zP * c_126[iRHS, iP]
                    s_126_n[iRHS, iP, iL, ippl] = C_126[iL] @ e_126_n[iRHS, iP]
                    s_126_m[iRHS, iP, iL, ippl] = C_126[iL] @ e_126_m[iRHS, iP]
                    s_45[iRHS, iP, iL, ippl, 0] = \
                        sfxz * ABDS[6, 6] * e_45[iRHS, iP, 0]
                    s_45[iRHS, iP, iL, ippl, 1] = \
                        sfyz * ABDS[7, 7] * e_45[iRHS, iP, 1]
                    e_45_new[iRHS, iP, iL, ippl, 0] = \
                        s_45[iRHS, iP, iL, ippl, 0] / C_45[iL, 0, 0]
                    e_45_new[iRHS, iP, iL, ippl, 1] = \
                        s_45[iRHS, iP, iL, ippl, 1] / C_45[iL, 1, 1]

    return s_126_n, s_126_m, s_45, e_126_n, e_126_m, e_45_new


@njit(nogil=True, parallel=True, cache=__cache)
def _pproc_Mindlin_3D_rgrid(ABDS: np.ndarray, coords2d: np.ndarray,
                            topo2d: np.ndarray, bounds: np.ndarray,
                            cell_per_layer: int, eshape: tuple,
                            sfx: np.ndarray, sfy: np.ndarray,
                            C_126: np.ndarray, C_45: np.ndarray,
                            e_126_n: np.ndarray, c_126: np.ndarray,
                            e_45: np.ndarray):
    nNEx, nNEy, nNEz = eshape
    nNE3d = nNEx * nNEy * nNEz
    nRHS = e_126_n.shape[0]
    nL = bounds.shape[0]
    nE2d, nNE2d = topo2d.shape
    nE3d = nE2d * nL * cell_per_layer
    nN3d = nE3d * nNE3d

    ftype = coords2d.dtype
    itype = topo2d.dtype
    s_126_n = np.zeros((nRHS, nN3d, 3), dtype=ftype)
    s_126_m = np.zeros((nRHS, nN3d, 3), dtype=ftype)
    s_45 = np.zeros((nRHS, nN3d, 2), dtype=ftype)
    e_126_m = np.zeros((nRHS, nN3d, 3), dtype=ftype)
    e_45_new = np.zeros((nRHS, nN3d, 2), dtype=ftype)
    coords3d = np.zeros((nN3d, 3), dtype=ftype)
    topo3d = np.zeros((nE3d, nNE3d), dtype=itype)

    # precalculate shear factors
    sfxz = np.zeros((nL, cell_per_layer, nNEz), dtype=ftype)
    sfyz = np.zeros((nL, cell_per_layer, nNEz), dtype=ftype)
    for iL in prange(nL):
        dZ = (bounds[iL, 1] - bounds[iL, 0]) / cell_per_layer
        ddZ = dZ / (nNEz - 1)
        for iCL in prange(cell_per_layer):
            for iNEz in prange(nNEz):
                zP = dZ * iCL + ddZ * iNEz
                sfxz[iL, iCL, iNEz], sfyz[iL, iCL, iNEz] = \
                    z_to_shear_factors(zP, sfx[iL, :], sfy[iL, :])

    for iL in prange(nL):
        dZ = (bounds[iL, 1] - bounds[iL, 0]) / cell_per_layer
        ddZ = dZ / (nNEz - 1)
        for iCL in prange(cell_per_layer):
            for iE2d in prange(nE2d):
                iE3d = nE2d * (cell_per_layer * iL + iCL) + iE2d
                for iNEx in prange(nNEx):
                    for iNEy in prange(nNEy):
                        iNE2d = iNEx * nNEy + iNEy
                        iN2d = topo2d[iE2d, iNE2d]
                        for iNEz in prange(nNEz):
                            iNE3d = iNEx * iNEy * nNEz + iNEz
                            iN3d = iE3d * nNE3d + iNE3d
                            topo3d[iE3d, iNE3d] = iN3d
                            coords3d[iN3d, 0] = coords2d[iN2d, 0]
                            coords3d[iN3d, 1] = coords2d[iN2d, 1]
                            coords3d[iN3d, 2] = dZ * iCL + ddZ * iNEz
                            for iRHS in prange(nRHS):
                                e_126_m[iRHS, iN3d] = \
                                    coords3d[iN3d, 2] * c_126[iRHS, iN2d]
                                s_126_n[iRHS, iN3d] = \
                                    C_126[iL] @ e_126_n[iRHS, iN2d]
                                s_126_m[iRHS, iN3d] = \
                                    C_126[iL] @ e_126_m[iRHS, iN3d]
                                s_45[iRHS, iN3d, 0] = \
                                    sfxz[iL, iCL, iNEz] * ABDS[6, 6] * \
                                    e_45[iRHS, iN2d, 0]
                                s_45[iRHS, iN3d, 1] = \
                                    sfyz[iL, iCL, iNEz] * ABDS[7, 7] * \
                                    e_45[iRHS, iN2d, 1]
                                e_45_new[iRHS, iN3d, 0] = \
                                    s_45[iRHS, iN3d, 0] / C_45[iL, 0, 0]
                                e_45_new[iRHS, iN3d, 1] = \
                                    s_45[iRHS, iN3d, 1] / C_45[iL, 1, 1]

    return s_126_n, s_126_m, s_45, e_126_n, e_126_m, e_45_new


def pproc_Mindlin_3D(ABDS: np.ndarray,
                     C_126: Iterable, C_45: Iterable,
                     bounds: Iterable, points: np.ndarray,
                     strains2d: np.ndarray, *args,
                     squeeze=True,
                     angles: Iterable = None, separate=True,
                     shear_factors: Iterable = None, **kwargs):
    # formatting
    ABDS = atleast3d(ABDS)
    strains2d = atleast4d(strains2d)
    if isinstance(C_126, np.ndarray):
        C_126 = atleast4d(C_126)
    else:
        C_126 = [atleast3d(C_126[i]) for i in range(len(C_126))]
    if isinstance(C_45, np.ndarray):
        C_45 = atleast4d(C_45)
    else:
        C_45 = [atleast3d(C_45[i]) for i in range(len(C_45))]
    if isinstance(bounds, np.ndarray):
        bounds = atleast3d(bounds)
    else:
        bounds = [atleast2d(bounds[i])
                  for i in range(len(bounds))]

    # transform material stiffness matrices to global
    if angles is not None:
        C_126_g = np.zeros_like(C_126)
        C_45_g = np.zeros_like(C_45)
        if isinstance(angles, np.ndarray):
            angles = atleast2d(angles)
        else:
            angles = [atleast1d(angles[i])
                      for i in range(len(angles))]
        for iLHS in range(len(angles)):
            C_126_g[iLHS], C_45_g[iLHS] = \
                material_stiffness_matrices(C_126[iLHS], C_45[iLHS],
                                            angles[iLHS])
        C_126 = C_126_g
        C_45 = C_45_g
        del C_126_g
        del C_45_g

    # calculate shear factors
    if shear_factors is None:
        shear_factors = []
        for i in range(len(C_126)):
            _, sf = shear_correction_data(ABDS[i], C_126[i], C_45[i],
                                          bounds[i])
            shear_factors.append(sf)
    else:
        if isinstance(shear_factors, np.ndarray):
            shear_factors = atleast4d(shear_factors)
        else:
            shear_factors = [atleast3d(shear_factors[i])
                             for i in range(len(shear_factors))]

    nLHS, nRHS, nP, _ = strains2d.shape
    e_126_n = ascont(strains2d[:, :, :, :3])
    s_126_n = np.zeros((nLHS, nRHS, nP, 3), dtype=ABDS.dtype)
    s_126_m = np.zeros((nLHS, nRHS, nP, 3), dtype=ABDS.dtype)
    s_45 = np.zeros((nLHS, nRHS, nP, 2), dtype=ABDS.dtype)
    e_126_m = np.zeros((nLHS, nRHS, nP, 3), dtype=ABDS.dtype)
    e_45 = np.zeros((nLHS, nRHS, nP, 2), dtype=ABDS.dtype)

    for i in range(nLHS):
        sfx_i = ascont(shear_factors[i][:, 0, :])
        sfy_i = ascont(shear_factors[i][:, 1, :])
        e_126_n_i = ascont(strains2d[i, :, :, :3])
        c_126_i = ascont(strains2d[i, :, :, 3:6])
        e_45_i = ascont(strains2d[i, :, :, 6:8])
        s_126_n[i], s_126_m[i], s_45[i], e_126_n[i], e_126_m[i], e_45[i] = \
            _pproc_Mindlin_3D(ABDS[i], sfx_i, sfy_i, C_126[i], C_45[i],
                              bounds[i], points, e_126_n_i, c_126_i, e_45_i)

    if separate:
        res3d = s_126_n, s_126_m, s_45, e_126_n, e_126_m, e_45
    else:
        res3d = s_126_n + s_126_m, s_45, e_126_n + e_126_m, e_45

    if squeeze:
        return tuple(map(np.squeeze, res3d))
    else:
        return res3d



# -*- coding: utf-8 -*-
from dewloosh.core.tools.kwargtools import getasany


standard_keys_ortho = ['E1', 'E2', 'E3',
                       'NU12', 'NU13', 'NU23', 'G12', 'G13', 'G23']
keys_ortho_all = standard_keys_ortho + ['NU21', 'NU32', 'NU31']
bulk_keys_ortho = ['E1', 'E2', 'E3', 'NU12',
                   'NU13', 'NU23', 'NU21', 'NU32', 'NU31']


class HookeError(Exception):
    ...


def group_mat_params(**params) -> tuple:
    """
    Groups and returns all input material parameters as a 3-tuple
    of dictionaries.

    Returns
    -------
    tuple
        3 dictionaries for Young's moduli, Poisson's ratios 
        and shear moduli.
    """
    PARAMS = {key.upper(): value for key, value in params.items()}
    keys = list(PARAMS.keys())
    E = {key: PARAMS[key] for key in keys if key[0] == 'E'}
    NU = {key: PARAMS[key] for key in keys if key[:2] == 'NU'}
    G = {key: PARAMS[key] for key in keys if key[0] == 'G'}
    return E, NU, G


def missing_std_params_ortho(**params):
    """Returns the ones missing from the 9 standard orthotropic parameters."""
    return [k for k in standard_keys_ortho if k not in params]


def missing_params_ortho(**params):
    """Returns the ones missing from the 12 orthotropic parameters."""
    return [k for k in keys_ortho_all if k not in params]


def has_std_params_ortho(**params) -> bool:
    """
    Returns True, if all 9 standard keys are provided.
    Standard keys are the 3 Young's moduli, the 3 shear moduli
    and the minor Poisson's ratios.
    """
    return len(missing_std_params_ortho(**params)) == 0


def has_all_params_ortho(**params) -> bool:
    """ Returns True, if all 12 keys of an orthotropic material are provided."""
    return len(missing_params_ortho(**params)) == 0


def get_iso_params(*args, **kwargs):
    """
    Returns all 12 orthotropic engineering constants for an 
    isotropic material. 
    Requires 2 independent constants to be provided.
    """
    KWARGS = {key.upper(): value for key, value in kwargs.items()}
    E = KWARGS.get('E', None)
    G = KWARGS.get('G', None)
    NU = KWARGS.get('NU', None)
    try:
        if E is None:
            E = 2*G*(1+NU)
        elif G is None:
            G = E/2/(1+NU)
        elif NU is None:
            NU = E/2/G - 1
    except TypeError:
        raise HookeError("At least 2 independent constants"
                         " must be defined for an isotropic material")
    params = \
        {'E1': E, 'E2': E, 'E3': E,
         'NU12': NU, 'NU13': NU, 'NU23': NU,
         'NU21': NU, 'NU31': NU, 'NU32': NU,
         'G12': G, 'G13': G, 'G23': G}
    return params


def get_triso_params(*args, **kwargs) -> dict:
    """
    Returns all 12 orthotropic engineering constants for a 
    transversely isotropic material. 
    From the total of 12 constants, exactly 5 must to be provided, 
    from which 1 must be the out-of-plane shear moduli. 
    The remaining 4 necessary constants can be provided in any combination.
    """
    KWARGS = {key.upper(): value for key, value in kwargs.items()}
    G12 = getasany(['G12', 'G21'], None, **KWARGS)
    G23 = getasany(['G23', 'G32'], None, **KWARGS)
    G13 = getasany(['G13', 'G23'], None, **KWARGS)
    E1 = KWARGS.get('E1', None)
    E2 = KWARGS.get('E2', None)
    E3 = KWARGS.get('E3', None)
    NU12 = KWARGS.get('NU12', None)
    NU21 = KWARGS.get('NU21', None)
    NU23 = KWARGS.get('NU23', None)
    NU32 = KWARGS.get('NU32', None)
    NU13 = KWARGS.get('NU13', None)
    NU31 = KWARGS.get('NU31', None)
    params = \
        {'E1': E1, 'E2': E2, 'E3': E3,
         'NU12': NU12, 'NU13': NU13, 'NU23': NU23,
         'NU21': NU21, 'NU31': NU31, 'NU32': NU32,
         'G12': G12, 'G13': G13, 'G23': G23}
    return params


def get_ortho_params(*args, **kwargs) -> dict:
    """
    Returns all 12 orthotropic engineering constants for an 
    orthotropic material. 
    From the total of 12 constants, 9 must to be provided, from which
    3 must be shear moduli. The remaining 6 necessary constants 
    can be provided in any combination.
    """
    KWARGS = {key.upper(): value for key, value in kwargs.items()}
    G12 = getasany(['G12', 'G21'], None, **KWARGS)
    assert G12 is not None, "Shear modulus G12 \
        must be provided as either G12 or G21!"
    G23 = getasany(['G23', 'G32'], None, **KWARGS)
    assert G23 is not None, "Shear modulus G23 \
        must be provided as either G23 or G32!"
    G13 = getasany(['G13', 'G23'], None, **KWARGS)
    assert G13 is not None, "Shear modulus G13 \
        must be provided as either G13 or G31!"
    E1 = KWARGS.get('E1', None)
    E2 = KWARGS.get('E2', None)
    E3 = KWARGS.get('E3', None)
    NU12 = KWARGS.get('NU12', None)
    NU21 = KWARGS.get('NU21', None)
    NU23 = KWARGS.get('NU23', None)
    NU32 = KWARGS.get('NU32', None)
    NU13 = KWARGS.get('NU13', None)
    NU31 = KWARGS.get('NU31', None)
    params = \
        {'E1': E1, 'E2': E2, 'E3': E3,
         'NU12': NU12, 'NU13': NU13, 'NU23': NU23,
         'NU21': NU21, 'NU31': NU31, 'NU32': NU32,
         'G12': G12, 'G13': G13, 'G23': G23}
    return params


if __name__ == '__main__':
    get_iso_params(E=1, NU=0.2)



# -*- coding: utf-8 -*-
from typing import Tuple, Union, Callable
import numpy as np
from numpy import ndarray
from numpy.linalg import inv
import sympy as sy

from ...material.hooke.utils import group_mat_params, get_iso_params
from ...material.hooke.sym import smat_sym_ortho_3d

from ..metashell import Surface, Layer


__all__ = ['MindlinShell']


class MindlinShellLayer(Layer):
    """
    Helper object for the stress analysis of a layer of a shell.
    """

    __loc__ = [-1., 0., 1.]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # locations of discrete points along the thickness
        self.zi = [self.loc_to_z(loc) for loc in MindlinShellLayer.__loc__]

        # Shear factors have to be multiplied by shear force to obtain shear
        # stress. They are determined externaly at discrete points, which
        # are later used for interpolation.
        self.shear_factors_x = np.array([0., 0., 0.])
        self.shear_factors_y = np.array([0., 0., 0.])

        # polinomial coefficients for shear factor interpoaltion
        self.sfx = None
        self.sfy = None

    def material_stiffness_matrices(self) -> Tuple[ndarray, ndarray]:
        """
        Returns and stores transformed material stiffness matrices.
        """
        Cm = self.hooke
        Cm_126 = Cm[0:3, 0:3]
        Cm_45 = Cm[3:, 3:]
        T_126, T_45 = self.rotation_matrices()
        R_126 = np.diag([1, 1, 2])
        R_45 = np.diag([2, 2])  
        C_126 = T_126 @ Cm_126 @ inv(R_126) @ T_126.T @ R_126
        C_45 = T_45 @ Cm_45 @ inv(R_45) @ T_45.T @ R_45    
        C_126[np.abs(C_126) < 1e-12] = 0.0
        C_45[np.abs(C_45) < 1e-12] = 0.0
        self.C_126 = C_126
        self.C_45 = C_45
        return C_126, C_45

    def rotation_matrices(self) -> Tuple[ndarray, ndarray]:
        """
        Produces transformation matrices T_126 and T_45.
        """
        T_126 = np.zeros([3, 3])
        T_45 = np.zeros([2, 2])
        angle = self.angle * np.pi/180
        #
        T_126[0, 0] = np.cos(angle)**2
        T_126[0, 1] = np.sin(angle)**2
        T_126[0, 2] = -np.sin(2*angle)
        T_126[1, 0] = T_126[0, 1]
        T_126[1, 1] = T_126[0, 0]
        T_126[1, 2] = -T_126[0, 2]
        T_126[2, 0] = np.cos(angle)*np.sin(angle)
        T_126[2, 1] = -T_126[2, 0]
        T_126[2, 2] = np.cos(angle)**2 - np.sin(angle)**2
        #
        T_45[0, 0] = np.cos(angle)
        T_45[0, 1] = -np.sin(angle)
        T_45[1, 1] = T_45[0, 0]
        T_45[1, 0] = -T_45[0, 1]
        #
        return T_126, T_45

    def stiffness_matrix(self) -> ndarray:
        """
        Returns the uncorrected stiffness contribution to the layer.
        """
        C_126, C_45 = self.material_stiffness_matrices()
        tmin = self.tmin
        tmax = self.tmax
        A = C_126*(tmax - tmin)
        B = (1/2)*C_126*(tmax**2 - tmin**2)
        D = (1/3)*C_126*(tmax**3 - tmin**3)
        S = C_45*(tmax - tmin)
        ABDS = np.zeros([8, 8])
        ABDS[0:3, 0:3] = A
        ABDS[0:3, 3:6] = B
        ABDS[3:6, 0:3] = B
        ABDS[3:6, 3:6] = D
        ABDS[6:8, 6:8] = S
        return ABDS

    def compile_shear_factors(self):
        """
        Prepares data for continuous interpolation of shear factors. Should
        be called if shear factors are already set.
        """
        coeff_inv = np.linalg.inv(np.array([[1, z, z**2] for z in self.zi]))
        self.sfx = np.matmul(coeff_inv, self.shear_factors_x)
        self.sfy = np.matmul(coeff_inv, self.shear_factors_y)

    def loc_to_shear_factors(self, loc: float):
        """
        Returns shear factor for local z direction by quadratic interpolation.
        Local coordinate is expected between -1 and 1.
        """
        z = self.loc_to_z(loc)
        monoms = np.array([1, z, z**2])
        return np.dot(monoms, self.sfx), np.dot(monoms, self.sfy)

    def approxfunc(self, data) -> Callable:
        z0, z1, z2 = self.zi
        z = np.array([[1, z0, z0**2], [1, z1, z1**2], [1, z2, z2**2]])
        a, b, c = np.linalg.inv(z) @ np.array(data)
        return lambda z: a + b*z + c*z**2


class MindlinPlateLayer(MindlinShellLayer):
    """
    Helper object for the stress analysis of a layer of a plate.
    """

    def stiffness_matrix(self) -> ndarray:
        """
        Returns the uncorrected stiffness contribution to the layer.
        """
        C_126, C_45 = self.material_stiffness_matrices()
        tmin = self.tmin
        tmax = self.tmax
        D = (1/3)*C_126*(tmax**3 - tmin**3)
        S = C_45*(tmax - tmin)
        ABDS = np.zeros([5, 5])
        ABDS[:3, :3] = D
        ABDS[3:, 3:] = S
        return ABDS


class MindlinShell(Surface):
    """
    Helper object for the stress analysis of a shell.
    
    Example
    -------
    >>> from sigmaepsilon.solid.model import MindlinShell as Model
    >>> model = {
    >>>     '0' : {
    >>>         'hooke' : Model.Hooke(E=2100000, nu=0.3),
    >>>         'angle' : 0., 
    >>>         'thickness' : 0.1
    >>>         },
    >>>     }
    >>> C = Model.from_dict(model).stiffness_matrix()
    
    """
    
    __layerclass__ = MindlinShellLayer

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    @classmethod    
    def Hooke(cls, *args, symbolic=False, **kwargs) -> Union[sy.Matrix, np.ndarray]:
        """
        Returns a Hooke matrix appropriate for shells.
        
        Parameters
        ----------
        symbolic : bool, Optional
            If Truem a symbolic matrix is returned.
        
        Returns
        -------
        Union[sympy.Matrix, numpy.ndarray]
            The matrix in symbolic or numeric form.
            
        Examples
        --------
        >>> from sigmaepsilon.solid.model import MindlinShell as Model
        >>> Model.Hooke(E=2100000, nu=0.3)
        
        >>> Model.Hooke(symbolic=True)
        
        """
        if symbolic:
            S = smat_sym_ortho_3d()
            S.row_del(2)
            S.col_del(2) 
            return S
        else:
            E, NU, G = group_mat_params(**kwargs)
            nE, nNU, nG = len(E), len(NU), len(G)
            nParams = sum([nE, nNU, nG])
            if nParams == 2:
                constants = get_iso_params(**E, **NU, **G)
            S = cls.Hooke(symbolic=True)
            subs = {s : constants[str(s)] for s in S.free_symbols}
            S = S.subs([(sym, val) for sym, val in subs.items()])
            S = np.array(S, dtype=float)
            return np.linalg.inv(S)

    def stiffness_matrix(self) -> ndarray:
        """
        Returns the stiffness matrix of the shell.
                
        Returns
        -------
        numpy.ndarray
            The ABDS matrix of the shell.
            
        """
        ABDS = super().stiffness_matrix()
        layers = self.layers()
        A11 = ABDS[0, 0]
        B11 = ABDS[0, 3]
        D11 = ABDS[3, 3]
        S55 = ABDS[6, 6]
        A22 = ABDS[1, 1]
        B22 = ABDS[1, 4]
        D22 = ABDS[4, 4]
        S44 = ABDS[7, 7]
        eta_x = 1/(A11*D11-B11**2)
        eta_y = 1/(A22*D22-B22**2)

        # Create shear factors. These need to be multiplied with the shear
        # force in order to obtain shear stress at a given height. Since the
        # shear stress distribution is of 2nd order, the factors are
        # determined at 3 locations per layer.
        for i, layer in enumerate(layers):
            zi = layer.zi
            Exi = layer.C_126[0, 0]
            Eyi = layer.C_126[1, 1]

            # first point through the thickness
            layer.shear_factors_x[0] = layers[i-1].shear_factors_x[-1]
            layer.shear_factors_y[0] = layers[i-1].shear_factors_y[-1]

            # second point through the thickness
            layer.shear_factors_x[1] = layer.shear_factors_x[0] - \
                eta_x*Exi*(0.5*(zi[1]**2-zi[0]**2)*A11-(zi[1]-zi[0])*B11)
            layer.shear_factors_y[1] = layer.shear_factors_y[0] - \
                eta_y*Eyi*(0.5*(zi[1]**2-zi[0]**2)*A22-(zi[1]-zi[0])*B22)

            # third point through the thickness
            layer.shear_factors_x[2] = layer.shear_factors_x[0] - \
                eta_x*Exi*(0.5*(zi[2]**2-zi[0]**2)*A11-(zi[2]-zi[0])*B11)
            layer.shear_factors_y[2] = layer.shear_factors_y[0] - \
                eta_y*Eyi*(0.5*(zi[2]**2-zi[0]**2)*A22-(zi[2]-zi[0])*B22)

        # remove numerical junk from the end
        layers[-1].shear_factors_x[-1] = 0.
        layers[-1].shear_factors_y[-1] = 0.

        # prepare data for interpolation of shear stresses in a layer
        for layer in layers:
            layer.compile_shear_factors()

        # potential energy using constant stress distribution
        # and unit shear force
        pot_c_x = 0.5/S55
        pot_c_y = 0.5/S44

        # positions and weights of Gauss-points
        gP = np.array([-np.sqrt(3/5), 0, np.sqrt(3/5)])
        gW = np.array([5/9, 8/9, 5/9])

        # potential energy using parabolic stress distribution
        # and unit shear force
        pot_p_x, pot_p_y = 0., 0.
        for layer in layers:
            dJ = 0.5*(layer.tmax - layer.tmin)
            Gxi = layer.C_45[0, 0]
            Gyi = layer.C_45[1, 1]
            for loc, weight in zip(gP, gW):
                sfx, sfy = layer.loc_to_shear_factors(loc)
                pot_p_x += 0.5*(sfx**2)*dJ*weight/Gxi
                pot_p_y += 0.5*(sfy**2)*dJ*weight/Gyi
        kx = pot_c_x/pot_p_x
        ky = pot_c_y/pot_p_y

        ABDS[6, 6] = kx * S55
        ABDS[7, 7] = ky * S44
        self.kx = kx
        self.ky = ky
        self.ABDS = ABDS
        self.SDBA = np.linalg.inv(ABDS)
        return ABDS


class MindlinPlate(MindlinShell):
    """
    Helper object for the stress analysis of a membrane.
    
    Example
    -------
    >>> from sigmaepsilon.solid.model import MindlinPlate as Model
    >>> model = {
    >>>     '0' : {
    >>>         'hooke' : Model.Hooke(E=2100000, nu=0.3),
    >>>         'angle' : 0., 
    >>>         'thickness' : 0.1
    >>>         },
    >>>     }
    >>> C = Model.from_dict(model).stiffness_matrix()
    
    """
    
    def stiffness_matrix(self) -> ndarray:
        """
        Returns the stiffness matrix of the plate.
                
        Returns
        -------
        numpy.ndarray
            The ABDS matrix of the plate.
            
        """
        ABDS_ = super().stiffness_matrix()
        ABDS = np.zeros([5, 5])
        ABDS[:3, :3] = ABDS_[3:6, 3:6]
        ABDS[3:, 3:] = ABDS_[6:8, 6:8]
        return ABDS



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from scipy.sparse import spmatrix
from scipy.sparse import coo_matrix as npcoo, csc_matrix as npcsc
import time
from typing import Union

from linkeddeepdict import LinkedDeepDict
from neumann.array import atleast2d, matrixform

from .mesh import FemMesh
from .utils import irows_icols_bulk
from .linsolve import solve_standard_form, unbox_lhs
from .imap import index_mappers, box_spmatrix, box_rhs, box_dof_numbering
from .dyn import (effective_modal_masses, Rayleigh_quotient,
                  natural_circular_frequencies)
from .constants import DEFAULT_MASS_PENALTY_RATIO


class Solver:
    """ 
    Base Class for solvers for the Finite Element Method.
    """


class StaticSolver(Solver):
    """
    A class to perform solutions for linear elastostatics
    and dynamics.

    Parameters
    ----------
    K : numpy.ndarray
        The stiffness matrix in in dense format.

    P : scipy.sparse.spmatrix
        The penalty stiffness matrix of the essential boundary conditions.

    f : numpy.ndarray
        The load vector as an 1d or 2d numpy array.

    gnum : numpy.ndarray
        Global dof numbering as 2d integer array.

    imap : Union[dict, numpy.ndarray], Optional
        Index mapper. Default is None.

    regular : bool, Optional
        If True it is assumed that the structure is regular. Default is True.

    mesh : FemMesh, Optional
        The mesh the input data belongs to. Only necessary for
        nonlinear calculations. Default is None.

    """

    def __init__(self, K: ndarray, P: spmatrix, f: ndarray, gnum: ndarray,
                 regular: bool = False, M: ndarray = None, mesh: FemMesh = None, **config):
        self.K_dense = K
        self.M_dense = M
        self.P_sparse = P
        self.gnum = gnum
        self.f = f
        self.u = None  # this is filled up during solution
        self.config = config
        self.regular = regular
        self.imap = None
        self.mesh = mesh
        self.regular = regular
        self.READY = False
        self.summary = []
        self._summary = LinkedDeepDict()
        self.core = self._box_()

    def update_stiffness(self, K_dense: ndarray):
        self.K_dense = K_dense
        self._preproc_(force=True)

    def _box_(self) -> 'StaticSolver':
        if self.imap is None and not self.regular:
            loc_to_glob, glob_to_loc = index_mappers(self.gnum,
                                                     return_inverse=True)
            gnum = box_dof_numbering(self.gnum, glob_to_loc)
            P_sparse = box_spmatrix(self.P_sparse, glob_to_loc)
            f = box_rhs(matrixform(self.f), loc_to_glob)
            #Kp, gnum, f, imap = box_fem_data_bulk(self.Kp, self.gnum, self.f)
            self.imap = glob_to_loc
            return StaticSolver(self.K_dense, P_sparse, f, gnum, regular=True)
        else:
            return self

    def _unbox_(self):
        assert not self.regular
        N = self.gnum.max() + 1
        self.u = unbox_lhs(self.core.u, self.imap, N=N)
        self.r = unbox_lhs(self.core.r, self.imap, N=N)

    def _preproc_(self, force: bool = False):
        if self.READY and not force:
            return
        _t0 = time.time()
        self.u = None
        self.N = self.gnum.max() + 1
        if self.config.get('sparsify', False):
            if not isinstance(self.f, npcsc):
                self.f = npcsc(self.f)
        self.krows, self.kcols = irows_icols_bulk(self.gnum)
        self.krows = self.krows.flatten()
        self.kcols = self.kcols.flatten()

        if not self.regular:
            self.core._preproc_()
            self.K_sparse = self.core.K_sparse
        else:
            self.K_sparse = npcoo((self.K_dense.flatten(), (self.krows, self.kcols)),
                            shape=(self.N, self.N), dtype=self.K_dense.dtype)
        _dt = time.time() - _t0
        self.READY = True
        self._summary['preproc', 'regular'] = self.regular
        self._summary['preproc', 'time'] = _dt
        self._summary['preproc', 'N'] = self.N

    def _proc_(self, preproc: bool = False, solver: str = None):
        if preproc:
            self._preproc_(force=True)
        _t0 = time.time()
        if not self.regular:
            self.core._proc_()
            _dt = time.time() - _t0
            self._summary['proc']['time'] = _dt
            return self
        K_sparse = self.K_sparse + self.P_sparse
        K_sparse.eliminate_zeros()
        K_sparse.sum_duplicates()
        self.u, summary = solve_standard_form(
            K_sparse, self.f, summary=True, solver=solver)
        self._summary['proc'] = summary

    def _postproc_(self):
        _t0 = time.time()
        if not self.regular:
            self.core._postproc_()
            self._summary['core'] = self.core._summary
            return self._unbox_()
        self.r = np.reshape(self.K_sparse.dot(self.u), self.f.shape) - self.f
        _dt = time.time() - _t0
        self._summary['postproc', 'time'] = _dt

    def solve(self, *, solver: str = None) -> ndarray:
        """
        Performs a linear elastostatic analysis and returns the unknown
        coefficients as a numpy array.

        Parameters
        ----------
        solver : str, Optional
            The solver to use. Currently supported options are 'scipy' 
            and 'pardiso'. If nothing is specified, we prefer 'pardiso' if 
            it is around, otherwise the solver falls back to SciPy.

        Returns
        -------
        numpy.ndarray
            The vector of nodal displacements as a numpy array.

        """
        self._summary = LinkedDeepDict()
        self._preproc_()
        self._proc_(solver=solver)
        self._postproc_()
        self.summary.append(self._summary)
        return self.u


class DynamicSolver(Solver):
    """
    A class to perform solutions for linear elastostatics
    and dynamics.

    Parameters
    ----------
    K : numpy.ndarray
        The stiffness matrix indense format.

    P : scipy.sparse.spmatrix
        The penalty stiffness matrix of the essential boundary conditions.

    gnum : numpy.ndarray
        Global dof numbering as 2d integer array.

    M : scipy.sparse.spmatrix, Optional
        The mass matrix. Default is None.

    regular : bool, Optional
        If True it is assumed that the structure is regular. Default is True.

    mesh : FemMesh, Optional
        The mesh the input data belongs to. Only necessary for
        nonlinear calculations. Default is None.

    """

    def __init__(self, K: ndarray, P: spmatrix, M: spmatrix, gnum: ndarray,
                 regular: bool = False, mesh: FemMesh = None,
                 penalty_ratio:float=DEFAULT_MASS_PENALTY_RATIO, **config):
        self.K_dense = K
        self.K_sparse = None
        self.M_sparse = M
        self.P_sparse = P
        self.gnum = gnum
        self.frequencies = None
        self.modal_shapes = None
        self.config = config
        self.regular = regular
        self.imap = None
        self.mesh = mesh
        self.regular = regular
        self.READY = False
        self.summary = []
        self._summary = LinkedDeepDict()
        self.penalty_ratio = penalty_ratio
        self.core = self._box_()

    def _box_(self) -> 'DynamicSolver':
        if self.imap is None and not self.regular:
            _, glob_to_loc = index_mappers(self.gnum, return_inverse=True)
            gnum = box_dof_numbering(self.gnum, glob_to_loc)
            P_sparse = box_spmatrix(self.P_sparse, glob_to_loc)
            M_sparse = box_spmatrix(self.M_sparse, glob_to_loc)
            self.imap = glob_to_loc
            return DynamicSolver(self.K_dense, P_sparse, M_sparse,
                                 gnum, regular=True, 
                                 penalty_ratio=self.penalty_ratio)
        else:
            return self

    def _unbox_(self):
        assert not self.regular
        N = self.gnum.max() + 1
        self.modal_shapes = unbox_lhs(self.core.modal_shapes, self.imap, N=N)
        self.frequencies = self.core.frequencies

    def _preproc_(self, force: bool = False):
        if self.READY and not force:
            return
        _t0 = time.time()
        self.N = self.gnum.max() + 1
        self.krows, self.kcols = irows_icols_bulk(self.gnum)
        self.krows = self.krows.flatten()
        self.kcols = self.kcols.flatten()

        if not self.regular:
            self.core._preproc_()
            self.K_sparse = self.core.K_sparse
        else:
            self.K_sparse = npcoo((self.K_dense.flatten(), (self.krows, self.kcols)),
                                  shape=(self.N, self.N), dtype=self.K_dense.dtype)
        _dt = time.time() - _t0
        self.READY = True
        self._summary['preproc', 'regular'] = self.regular
        self._summary['preproc', 'time'] = _dt
        self._summary['preproc', 'N'] = self.N

    def _proc_free_(self, preproc: bool = False, **kwargs):
        if preproc:
            self._preproc_(force=True)
        _t0 = time.time()
        if not self.regular:
            self.core._proc_free_(**kwargs)
            _dt = time.time() - _t0
            self._summary['proc-free']['time'] = _dt
            return self
        K_sparse = self.K_sparse + self.P_sparse
        M_sparse = self.M_sparse + self.P_sparse * self.penalty_ratio
        K_sparse.eliminate_zeros()
        K_sparse.sum_duplicates()
        freqs, shapes = natural_circular_frequencies(
            K_sparse, M_sparse, return_vectors=True, **kwargs)
        self.frequencies = freqs
        self.modal_shapes = shapes
        
    def _postproc_(self):
        _t0 = time.time()
        if not self.regular:
            self.core._postproc_()
            self._summary['core'] = self.core._summary
            return self._unbox_()
        _dt = time.time() - _t0
        self._summary['postproc', 'time'] = _dt

    def natural_circular_frequencies(self, *, return_vectors: bool = False, 
                                     **kwargs):
        """
        Returns natural circular frequencies. The call forwards all parameters
        to :func:`sigmaepslion.solid.fem.dyn.natural_circular_frequencies`, 
        see the docs there for the details. 

        Parameters
        ----------
        return_vectors : bool, Optional
            To return eigenvectors or not. Default is False.
        penalty_ratio : float, Optional
            The penalty ratio. Only if 'penalize' is True. 
            Default is `sigmaepsilon.solid.fem.config.DEFAULT_PENALTY_RATIO`.

        See also
        --------
        :func:`sigmaepslion.solid.fem.dyn.natural_circular_frequencies`

        Returns
        -------
        numpy.ndarray
            The natural circular frequencies.
        numpy.ndarray
            The eigenvectors, only if 'return_vectors' is True.

        """
        if self.frequencies is None:
            self._preproc_()
            self._proc_free_(**kwargs)
            self._postproc_()
        if return_vectors:
            return self.frequencies, self.modal_shapes
        return self.frequencies



# -*- coding: utf-8 -*-
from numba import njit, prange
import numpy as np
from numpy import ndarray

from ...material.utils import HMH_M
from .surface import Surface


__cache = True


__all__ = ['Membrane', 'DrillingMembrane']


_NSTRE_ = 3
_NDOFN_ = 2
_NHOOKE_ = 3


@njit(nogil=True, parallel=True, cache=__cache)
def strain_displacement_matrix(dshp: ndarray, jac: ndarray):
    nE = jac.shape[0]
    nP, nN = dshp.shape[:2]
    nTOTV = nN * _NDOFN_
    B = np.zeros((nE, nP, _NSTRE_, nTOTV), dtype=dshp.dtype)
    for iE in prange(nE):
        for iP in prange(nP):
            gdshp = dshp[iP] @ np.linalg.inv(jac[iE, iP])
            for i in prange(nN):
                B[iE, iP, 0, 0 + i*_NDOFN_] = gdshp[i, 0]
                B[iE, iP, 1, 1 + i*_NDOFN_] = gdshp[i, 1]
                B[iE, iP, 2, 0 + i*_NDOFN_] = gdshp[i, 1]
                B[iE, iP, 2, 1 + i*_NDOFN_] = gdshp[i, 0]
    return B


@njit(nogil=True, parallel=True, cache=__cache)
def material_strains(strns: ndarray, z: float):
    nE, nP = strns.shape[:2]
    res = np.zeros((nE, nP, _NHOOKE_), dtype=strns.dtype)
    res[:, :, :3] = strns[:, :, :3]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def HMH(estrs: np.ndarray):
    nE, nP = estrs.shape[:2]
    res = np.zeros((nE, nP), dtype=estrs.dtype)
    for iE in prange(nE):
        for jNE in prange(nP):
            res[iE, jNE] = HMH_M(estrs[iE, jNE])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def model_stiffness_iso_homg(C: ndarray, t: ndarray):
    res = np.zeros_like(C)
    for i in prange(res.shape[0]):
        res[i, :, :] = C[i, :, :] * t[i]
    return res


class Membrane(Surface):
    
    dofs = ('UX', 'UY')
    strn = ('exx', 'eyy', 'exy')
    
    NDOFN = _NDOFN_
    NSTRE = _NSTRE_

    @classmethod
    def model_stiffness_matrix_iso_homg(cls, C, t, *args, **kwargs):
        return model_stiffness_iso_homg(C, t)

    @classmethod
    def strain_displacement_matrix(cls, *args, dshp=None, jac=None, **kwargs):
        return strain_displacement_matrix(dshp, jac)

    @classmethod
    def HMH(cls, data, *args, **kwargs):
        return HMH(data)

    @classmethod
    def material_strains(cls, model_strains, z, t, *args, **kwargs):
        return material_strains(model_strains, z)
    

class DrillingMembrane(Membrane):
    
    dofs = ('UX', 'UY', 'ROTZ')

    NDOFN = 3
    NSTRE = _NSTRE_
    


# -*- coding: utf-8 -*-
from neumann.array import atleast2d, atleastnd

from ..utils import topo_to_gnum

from .solid import Solid
from .utils import stresses_from_strains


class Surface(Solid):
    
    strn = ('exx', 'eyy', 'exy', 'kxx', 'kyy', 'kxy', 'exz', 'eyz')

    def model_stiffness_matrix(self, *args, **kwargs):
        C = self.material_stiffness_matrix()
        t = self.thickness()
        return self.model_stiffness_matrix_iso_homg(C, t)

    @classmethod
    def model_stiffness_matrix_iso_homg(cls, *args,  **kwargs):
        raise NotImplementedError

    @classmethod
    def material_strains(cls, *args, **kwargs):
        raise NotImplementedError

    def ABDS(self, *args, **kwargs):
        return self.model_stiffness_matrix(*args, **kwargs)

    def strains_at(self, lcoords, *args,  z=None, topo=None, **kwargs):
        if topo is None:
            topo = self.topology().to_numpy()
        lcoords = atleast2d(lcoords)
        dshp = self.shape_function_derivatives(lcoords)
        ecoords = self.local_coordinates(topo=topo)
        jac = self.jacobian_matrix(dshp, ecoords)
        gnum = topo_to_gnum(topo, self.NDOFN)
        
        
        dofsol1d = self.pointdata.dofsol.to_numpy().flatten()
        
        dofsol = self.pointdata.dofsol
        dofsol = atleastnd(dofsol, 3, back=True)
        nP, nDOF, nRHS = dofsol.shape
        dofsol = dofsol.reshape(nP * nDOF, nRHS)
        
        
        B = self.strain_displacement_matrix(dshp, jac)
        if z is None:
            # return generalized model strains
            return self.model_strains(dofsol1d, gnum, B)
        else:
            # returns material strains
            t = self.thickness()
            model_strains = self.model_strains(dofsol1d, gnum, B)
            return self.material_strains(model_strains, z, t)

    def stresses_at(self, *args, z=None, topo=None, **kwargs):
        """
        Returns stresses for every node of every element.
        """
        # The next code line returns either generalized strains
        # or material strains, depending on the value of z
        strns = self.strains_at(*args, z=z, topo=topo, **kwargs)
        if z is None:
            C = self.model_stiffness_matrix()
        else:
            C = self.material_stiffness_matrix()
        if 'HMH' in args:
            return self.HMH(stresses_from_strains(C, strns))
        return stresses_from_strains(C, strns)



# -*- coding: utf-8 -*-
import numpy as np
from numpy import swapaxes as swap
from typing import Tuple, Union, Iterable

from linkeddeepdict import LinkedDeepDict

from .problem import NavierProblem
from .loads import LoadGroup, NavierLoadError
from .preproc import lhs_Navier
from .postproc import postproc
from .proc import linsolve


class RectangularPlate(NavierProblem):
    """
    A class to handle semi-analytic solutions of rectangular plates with
    specific boudary conditions.

    Parameters
    ----------
    size : Tuple[float]
        The size of the rectangle.

    shape : Tuple[int]
        Numbers of harmonic terms involved in both directions.

    """

    def __init__(self, size: Tuple[float], shape: Tuple[int], *,
                 D11: float = None, D12: float = None, D22: float = None,
                 D66: float = None, S44: float = None, S55: float = None):
        super().__init__()
        self.size = np.array(size, dtype=float)
        self.shape = np.array(shape, dtype=int)
        self.D11 = D11
        self.D12 = D12
        self.D22 = D22
        self.D66 = D66
        self.S44 = S44
        self.S55 = S55

    def solve(self, loads: Union[dict, LoadGroup], points: Iterable):
        """
        Solves the problem and calculates all entities at the specified points.

        Parameters
        ----------
        loads : Union[dict, LoadGroup]
            The loads.

        points : Iterable
            2d float array of coordinates, where the results are to be evaluated.

        Returns
        -------
        dict
            A dictionary with a same layout as the input.

        """
        # STIFFNESS
        D = np.zeros((3, 3))
        D[0, 0] = self.D11
        D[0, 1] = self.D12
        D[1, 0] = self.D12
        D[1, 1] = self.D22
        D[2, 2] = self.D66
        if self.S44 is not None:
            S = np.zeros((2, 2))
            S[0, 0] = self.S44
            S[1, 1] = self.S55
        else:
            S = None
        LHS = lhs_Navier(self.size, self.shape, D, S, squeeze=False)

        # LOADS
        if isinstance(loads, LoadGroup):
            _loads = loads
        elif isinstance(loads, dict):
            _loads = LoadGroup.from_dict(loads)
        else:
            raise NavierLoadError()
        _loads.problem = self
        LC = list(_loads.cases())
        RHS = np.stack(list(lc.rhs() for lc in LC))
        
        # SOLUTION
        points = np.array(points)
        coeffs = linsolve(LHS, RHS, squeeze=False)
        res = postproc(self.size, self.shape, points, coeffs, 
                       RHS, D, S, squeeze=False)
        # (nRHS, nLHS, nP, nX)
        res = swap(res, 1, 2)
        # (nRHS, nP, nLHS, nX)
        result = LinkedDeepDict()
        for i, lc in enumerate(_loads.cases()):
            result[lc.address] = np.squeeze(res[i, :, :])
        result.lock()
        return result



# -*- coding: utf-8 -*-
import numpy as np

from neumann import squeeze
from neumann.linalg.sparse.csr import csr_matrix as csr

from sigmaepsilon.solid.fem.mesh import FemMesh as Mesh

__all__ = ['FemMesh']


class FemMesh(Mesh):
    ...
    
    
class FuckedUpFemMesh(Mesh):
    ...
    
    @squeeze(True)
    def load_vector(self, *args, **kwargs):
        raise NotImplementedError
        # pointdata
        nodal_loads = self.root().pointdata.loads.to_numpy()
        nodal_loads = atleast3d(nodal_loads, back=True)  # (nP, nDOF, nRHS)
        f_p = fem_load_vector(values=nodal_loads, squeeze=False)
               
        # celldata
        #blocks = self.cellblocks(inclusive=True)
        #def foo(b): return b.celldata.body_load_vector(squeeze=False, shape=shp)
        #f_c = np.sum(list(map(foo, blocks)))
        """
        if not self.is_compatible():
            # distribute nodal loads
            blocks = list(self.cellblocks(inclusive=True))
            def foo(b): return b.celldata.distribute_nodal_data(loads, 'loads')
            list(map(foo, blocks))
            # collect nodal loads
            N = len(loads)
            def foo(b): return b.celldata.collect_nodal_data('loads', N=N)
            loads = np.sum(list(map(foo, blocks)), axis=0)
        """
        return f_p
    
    def penalty_matrix_coo(self, *args, eliminate_zeros=True,
                           sum_duplicates=True, ensure_comp=False, 
                           distribute=False, **kwargs):
        raise NotImplementedError
        """A penalty matrix that enforces essential(Dirichlet) 
        boundary conditions. Returns a scipy sparse matrix in coo format."""
        
        # essential boundary conditions
        fixity = self.root().pointdata.fixity.to_numpy()
        K_coo = fem_penalty_matrix_coo(values=fixity, eliminate_zeros=eliminate_zeros, 
                                       sum_duplicates=sum_duplicates)
        
        if distribute:
            # distribute nodal fixity
            fixity = self.root().pointdata.fixity.to_numpy().astype(int)
            blocks = list(self.cellblocks(inclusive=True))
            def foo(b): return b.celldata.distribute_nodal_data(fixity, 'fixity')
            list(map(foo, blocks))
            # assemble
            def foo(b): return b.celldata.penalty_matrix_coo()
            K_coo = np.sum(list(map(foo, blocks))).tocoo()
        
        if ensure_comp and not self.is_compatible():
            # penalty arising from incompatibility
            p = kwargs.get('compatibility_penalty', None)
            if p is not None:
                K_coo += self.compatibility_penalty_matrix_coo(eliminate_zeros=False, p=p)
                
        return K_coo.tocoo()

    def compatibility_penalty_matrix_coo(self, *args, eliminate_zeros=True, p=1e12, **kwargs):
        blocks = self.cellblocks(inclusive=True)
        nam_csr_tot = csr(self.nodal_approximation_matrix_coo(eliminate_zeros=False))
        def foo(b): return b.celldata.compatibility_penalty_matrix_coo(nam_csr_tot=nam_csr_tot, p=p)
        res = np.sum(list(map(foo, blocks))).tocoo()
        if eliminate_zeros:
            res.eliminate_zeros()
        return res

    def is_compatible(self):
        blocks = self.cellblocks(inclusive=True)
        def fltr(b): return not b.celldata.compatible
        return len(list(filter(fltr, blocks))) == 0

if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
from typing import Union
import numpy as np
import sympy as sy
from numpy import ndarray
from numpy.linalg import inv

from ..material.hooke.utils import group_mat_params, get_iso_params
from ..material.hooke.sym import smat_sym_ortho_3d

from .metashell import Surface, Layer


__all__ = ['Membrane', 'MembraneLayer']


class MembraneLayer(Layer):
    """
    Helper object for the stress analysis of a layer of a membrane.
    """

    __loc__ = [-1., 1.]  # locations of approximation points
    __shape__ = (3, 3)  # shape of the ABDS matrix

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # locations of discrete points along the thickness
        self.zi = [self.loc_to_z(loc) for loc in self.__loc__]

    def material_stiffness_matrix(self) -> ndarray:
        """
        Returns and stores the transformed material stiffness matrix.
        """
        Cm_126 = self.hooke
        T_126 = self.rotation_matrix()
        R_126 = np.diag([1, 1, 2])
        C_126 = T_126 @ Cm_126 @ inv(R_126) @ T_126.T @ R_126
        C_126[np.abs(C_126) < 1e-12] = 0.0
        self.C_126 = C_126
        return C_126

    def rotation_matrix(self) -> ndarray:
        """
        Returns the transformation matrix.
        """
        angle = self.angle * np.pi / 180
        T_126 = np.zeros([3, 3])
        T_126[0, 0] = np.cos(angle)**2
        T_126[0, 1] = np.sin(angle)**2
        T_126[0, 2] = -np.sin(2 * angle)
        T_126[1, 0] = T_126[0, 1]
        T_126[1, 1] = T_126[0, 0]
        T_126[1, 2] = -T_126[0, 2]
        T_126[2, 0] = np.cos(angle) * np.sin(angle)
        T_126[2, 1] = -T_126[2, 0]
        T_126[2, 2] = np.cos(angle)**2 - np.sin(angle)**2
        return T_126

    def stiffness_matrix(self) -> ndarray:
        """
        Returns the stiffness contribution of the layer.
        """
        return self.material_stiffness_matrix() * (self.tmax - self.tmin)

    def approxfunc(self, data):
        """
        Returns a function that can be used for approximations thorugh
        the thickness.
        """
        z0, z1 = self.zi
        z = np.array([[1, z0], [1, z1]])
        a, b = np.linalg.inv(z) @ np.array(data)
        return lambda z: a + b*z


class Membrane(Surface):
    """
    Helper object for the stress analysis of a membrane.
    
    Example
    -------
    >>> from sigmaepsilon.solid.model import Membrane as Model
    >>> model = {
    >>>     '0' : {
    >>>         'hooke' : Model.Hooke(E=2100000, nu=0.3),
    >>>         'angle' : 0., 
    >>>         'thickness' : 0.1
    >>>         },
    >>>     }
    >>> C = Model.from_dict(model).stiffness_matrix()
    
    """
    
    __layerclass__ = MembraneLayer
    
    __imap__ = {0 : (0, 0), 1 : (1, 1), 2 : (2, 2),
                5 : (1, 2), 4 : (0, 2), 3 : (0, 1)}
        
    @classmethod    
    def Hooke(cls, *args, symbolic=False, **kwargs) -> Union[sy.Matrix, np.ndarray]:
        """
        Returns a Hooke matrix appropriate for membranes.
        
        Parameters
        ----------
        symbolic : bool, Optional
            If Truem a symbolic matrix is returned.
        
        Returns
        -------
        Union[sy.Matrix, np.ndarray]
            The matrix in symbolic or numeric form.
            
        Examples
        --------
        >>> from sigmaepsilon.solid.model import Membrane as Model
        >>> Model.Hooke(E=2100000, nu=0.3)
        
        >>> Model.Hooke(symbolic=True)
            
        """
        if symbolic:
            S = smat_sym_ortho_3d()
            S.row_del(2)
            S.row_del(2)
            S.row_del(2)
            S.col_del(2)
            S.col_del(2)
            S.col_del(2)  
            return S
        else:
            E, NU, G = group_mat_params(**kwargs)
            nE, nNU, nG = len(E), len(NU), len(G)
            nParams = sum([nE, nNU, nG])
            if nParams == 2:
                constants = get_iso_params(**E, **NU, **G)
            S = cls.Hooke(symbolic=True)
            subs = {s : constants[str(s)] for s in S.free_symbols}
            S = S.subs([(sym, val) for sym, val in subs.items()])
            S = np.array(S, dtype=float)
            return np.linalg.inv(S)
            
    def stiffness_matrix(self) -> ndarray:
        """
        Returns the stiffness matrix of the membrane.
                
        Returns
        -------
        ndarray
            The ABDS matrix of the membrane.
            
        """
        self.ABDS = super().stiffness_matrix()
        self.SDBA = np.linalg.inv(self.ABDS)
        return self.ABDS



# -*- coding: utf-8 -*-
from numba import njit, prange
import numpy as np
from numpy import ndarray

from ...material.utils import HMH_S

from .surface import Surface


__cache = True


__all__ = ['Plate']


_NSTRE_ = 5
_NDOFN_ = 3
_NHOOKE_ = 5


@njit(nogil=True, parallel=True, cache=__cache)
def strain_displacement_matrix(dshp: ndarray, jac: ndarray):
    nE = jac.shape[0]
    nP, nN = dshp.shape[:2]
    nTOTV = nN * _NDOFN_
    B = np.zeros((nE, nP, _NSTRE_, nTOTV), dtype=dshp.dtype)
    for iE in prange(nE):
        for iP in prange(nP):
            gdshp = dshp[iP] @ np.linalg.inv(jac[iE, iP])
            for i in prange(nN):
                B[iE, iP, 0, 1 + i*_NDOFN_] = -gdshp[i, 0]
                B[iE, iP, 1, 2 + i*_NDOFN_] = -gdshp[i, 1]
                B[iE, iP, 2, 1 + i*_NDOFN_] = -gdshp[i, 1]
                B[iE, iP, 2, 2 + i*_NDOFN_] = -gdshp[i, 0]
                B[iE, iP, 3, 0 + i*_NDOFN_] = gdshp[i, 0]
                B[iE, iP, 3, 1 + i*_NDOFN_] = -1
                B[iE, iP, 4, 0 + i*_NDOFN_] = gdshp[i, 1]
                B[iE, iP, 4, 2 + i*_NDOFN_] = -1
    return B


@njit(nogil=True, parallel=True, cache=__cache)
def material_strains(model_strains: ndarray, z: float, t: float):
    nE, nP = model_strains.shape[:2]
    res = np.zeros((nE, nP, _NHOOKE_), dtype=model_strains.dtype)
    res[:, :, :3] = model_strains[:, :, :3] * z
    for i in prange(nE):
        res[i, :, 3:] = (5/4) * (1 - 4 * (z / t[i])**2) * \
            model_strains[i, :, 3:]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def HMH(estrs: np.ndarray):
    nE, nP = estrs.shape[:2]
    res = np.zeros((nE, nP), dtype=estrs.dtype)
    for iE in prange(nE):
        for jNE in prange(nP):
            res[iE, jNE] = HMH_S(estrs[iE, jNE])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def model_stiffness_iso_homg(C: ndarray, t: ndarray):
    res = np.zeros_like(C)
    for i in prange(res.shape[0]):
        res[i, :3, :3] = C[i, :3, :3] * (t[i]**3 / 12)
        res[i, 3:, 3:] = C[i, 3:, 3:] * (t[i] * 5 / 6)
    return res


class Plate(Surface):
    
    dofs = ('UZ', 'ROTX', 'ROTY')
    strn = ('kxx', 'kyy', 'kxy' 'exz', 'eyz')
    
    NDOFN = _NDOFN_
    NSTRE = _NSTRE_

    @classmethod
    def model_stiffness_matrix_iso_homg(cls, C, t, *args, **kwargs):
        return model_stiffness_iso_homg(C, t)

    @classmethod
    def strain_displacement_matrix(cls, *args, dshp=None, jac=None, **kwargs):
        return strain_displacement_matrix(dshp, jac)

    @classmethod
    def HMH(cls, data, *args, **kwargs):
        return HMH(data)

    @classmethod
    def material_strains(cls, model_strains, z, t, *args, **kwargs):
        return material_strains(model_strains, z, t)
